"""example Core App Generic Views Base."""

from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect
from django.views.generic.base import TemplateResponseMixin, TemplateView


class HtmxTemplateResponseMixin(TemplateResponseMixin):
    """Extends Django TemplateResponseMixin with HTMX functionality.

    Renders an htmx template or a generic template response.

    MIT License

    Copyright (c) 2021 Pablo Rivera

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """

    htmx_template_name = None

    def get_template_names(self):
        """Return a list of template names.

        May not be called if render_to_response() is overridden.
        """
        super().get_template_names()

        if self.template_name is None and self.htmx_template_name is None:
            raise ImproperlyConfigured(
                "HtmxTemplateResponseMixin requires either a definition of "
                "'htmx_template_name' and 'template_name' or an "
                "implementation of 'get_template_names()'"
            )

        if self.request.htmx:
            return [self.htmx_template_name]

        return [self.template_name]


class HtmxTemplateView(HtmxTemplateResponseMixin, TemplateView):
    """Extends Django Template View with HTMX functionality.

    Renders an htmx template or a generic template response.
    """

    def get(self, request, *args, **kwargs):
        """Extends GET to check if the user has defined the
        'not_htmx_request_redirect' attribute. If they have
        and the request is not using HTMX then redirect the
        user to the value of the attribute.
        """
        if hasattr(self, "not_htmx_request_redirect"):
            if request.htmx:
                return super().get(self, request, *args, **kwargs)
            else:
                return redirect(self.not_htmx_request_redirect)
        return super().get(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """Extends POST to check if the user has defined the
        'not_htmx_request_redirect' attribute. If they have
        and the request is not using HTMX then redirect the
        user to the value of the attribute.
        """
        if hasattr(self, "not_htmx_request_redirect"):
            if request.htmx:
                return super().post(self, request, *args, **kwargs)
            else:
                return redirect(self.not_htmx_request_redirect)
        return super().post(self, request, *args, **kwargs)



"""Test tags tagged field utils."""

# import subprocess
# from io import StringIO
# from contextlib import redirect_stdout

from io import StringIO

# import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management import call_command

# from django.contrib.contenttypes.models import ContentType
# from django.core import management
# from django.core.exceptions import ValidationError
# from django.test import override_settings
# from hypothesis import given
# from hypothesis import strategies as st
from hypothesis.extra.django import TestCase

from tag_me.models import TaggedFieldModel, UserTag
from tag_me.utils.helpers import (  # update_models_with_tagged_fields_table,
    get_model_content_type,
    get_model_tagged_fields_choices,
    get_model_tagged_fields_field_and_verbose,
    get_models_with_tagged_fields,
    get_models_with_tagged_fields_choices,
    get_user_field_choices_as_list_tuples,
)
from tag_me.utils.tag_mgmt_system import (
    update_models_with_tagged_fields_table,
)

User = get_user_model()


class TestTagHelpers(TestCase):
    """Test the tags helper
    Uses `tests` models to make these tests.  This way we can have an
    explicit test set up.
    """

    def setUp(self):
        # make sure all migrations are up to date
        self.out_makemigration = StringIO()
        import logging

        logger = logging.getLogger(__name__)
        call_command(
            "makemigrations",
            stdout=self.out_makemigration,
        )
        logger.info(
            f"----------- MAKE MIGRATION: {self.out_makemigration.getvalue()}"
        )

        # Run migrations so the tags management commads are run
        self.out_migration = StringIO()
        call_command(
            "migrate",
            stdout=self.out_migration,
        )
        logger.info(f"----------- MIGRATION: {self.out_migration.getvalue()}")

        _model = "TaggedFieldTestModel"
        _field_1 = "tagged_field_1"
        _field_2 = "tagged_field_2"

        self.tag_string1 = '"apple ball cat'
        self.tag_string1_result = [
            "apple",
            "ball",
            "cat",
        ]
        self.tag_string2 = "tree, flower cactus"

        self.tag_string3 = "car, truck plane"

        self.user1 = User.objects.create(
            username="user1",
            password="pw_user1",
            email="user1@email.com",
        )
        self.user2 = User.objects.create(
            username="user2",
            password="pw_user2",
            email="user2@email.com",
        )
        self.user3 = User.objects.create(
            username="user3",
            password="pw_user3",
            email="user3@email.com",
        )
        # # Add all the users to the UserTag table.
        self.out_add_user_tags = StringIO()
        call_command(
            "add_user_tags",
            stdout=self.out_add_user_tags,
        )

        logger.info(
            f"----------- ADD USER ADD USER TAGS: {self.out_add_user_tags.getvalue()}"
        )

        self.model_1_field_1 = TaggedFieldModel.objects.get(
            model_name=_model,
            field_name=_field_1,
        )
        self.model_1_field_2 = TaggedFieldModel.objects.get(
            model_name=_model,
            field_name=_field_2,
        )

        self.user1_tag1 = UserTag.objects.get(
            tagged_field=self.model_1_field_1,
            user=self.user1,
        )
        self.user1_tag1.tags = "User1 Tag1"
        self.user1_tag1.save()

        self.user2_tag1 = UserTag.objects.get(
            tagged_field=self.model_1_field_1,
            user=self.user2,
        )
        self.user2_tag1.tags = "User2 Tag1"
        self.user2_tag1.save()

        self.user3_tag1 = UserTag.objects.get(
            tagged_field=self.model_1_field_1,
            user=self.user3,
        )
        self.user3_tag1.tags = "User3 Tag1"
        self.user3_tag1.save()

    def test_get_model_tagged_fields_field_and_verbose_empty_verbose(self):
        fields = get_model_tagged_fields_field_and_verbose(
            model_verbose_name="",
        )
        assert fields == [(None, None)]

    def test_get_models_with_tagged_fields_apps_in_INSTALLED(self):
        with self.settings(PROJECT_APPS=None):
            models = get_models_with_tagged_fields()

            assert "<class 'tests.models.Post'>" in str(models)
            assert "<class 'tests.models.TaggedFieldTestModel'>" in str(models)

    def test_get_models_with_tagged_fields_apps_in_PROJECT(self):
        with self.settings(PROJECT_APPS=settings.INSTALLED_APPS):
            models = get_models_with_tagged_fields()

            assert "<class 'tests.models.Post'>" in str(models)
            assert "<class 'tests.models.TaggedFieldTestModel'>" in str(models)

    def test_get_models_with_tagged_fields(self):
        fields = get_models_with_tagged_fields()

        assert "<class 'tests.models.TaggedFieldTestModel'>" in str(fields)
        assert "<class 'tests.models.Post'>" in str(fields)

    def test_get_models_with_tagged_fields_choices(self):
        choices = get_models_with_tagged_fields_choices()

        assert "('Tagged Field Test Model', 'Tagged Field Test Model')" in str(
            choices
        )

    def test_get_model_tagged_fields_choices_with_feature_name(self):
        choices1 = get_model_tagged_fields_choices(
            feature_name="Tagged Field Test Model"
        )

        assert "('Tagged Field 1', 'Tagged Field 1')" in str(choices1)

        assert "('Tagged Field 2', 'Tagged Field 2')" in str(choices1)

    def test_get_model_tagged_fields_choices_without_feature_name(self):
        choices1 = get_model_tagged_fields_choices(feature_name="")

        assert choices1 == [(None, None)]

    def test_get_model_content_type_with_verbose(self):
        content = get_model_content_type(
            model_verbose_name="Tagged Field Test Model",
        )

        assert content.app_label == "tests"
        assert content.model == "taggedfieldtestmodel"

    def test_get_model_content_type_with_out_verbose(self):
        content = get_model_content_type(
            model_verbose_name="",
        )

        assert content is None

    def test_get_user_field_choices_as_list_tuples(self):
        choices_all = UserTag.objects.all()
        choices_1 = get_user_field_choices_as_list_tuples(
            model_verbose_name="Tagged Field Test Model",
            field_name="tagged_field_1",
            user=self.user1,
        )
        choices_2 = get_user_field_choices_as_list_tuples(
            model_verbose_name="Tagged Field Test Model",
            field_name="tagged_field_1",
            user=self.user2,
        )

        choices_3 = get_user_field_choices_as_list_tuples(
            model_verbose_name="Tagged Field Test Model",
            field_name="tagged_field_1",
            user=self.user3,
        )

        assert self.user1_tag1 in choices_all and choices_1
        assert self.user2_tag1 in choices_all and choices_2
        assert self.user3_tag1 in choices_all and choices_3
        #
        assert self.user1_tag1 not in choices_2 and choices_3
        assert self.user2_tag1 not in choices_1 and choices_3
        assert self.user3_tag1 not in choices_1 and choices_2

    def test_tagged_field_models_table_populated_ok(self):
        TaggedFieldModel.objects.all().delete()
        assert not TaggedFieldModel.objects.all().exists()
        update_models_with_tagged_fields_table()
        assert TaggedFieldModel.objects.all().exists()



"""Users need to be added to the UserTag table."""

import logging

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import (
    DatabaseError,
    DataError,
    IntegrityError,
    transaction,
)

from tag_me.models import (
    TagBase,
    TaggedFieldModel,
    TagMeSynchronise,
    UserTag,
)
from tag_me.utils.helpers import (
    get_model_tagged_fields_field_and_verbose,
    get_models_with_tagged_fields,
    stdout_with_optional_color,
)

logger = logging.getLogger(__name__)

User = get_user_model()


def generate_user_tag_table_records(
    user=None,
):
    """
    Generates `UserTag` records for users who are not yet present in the table.

    This function automatically creates `UserTag` entries for each combination of:

    * **Existing Users (if no user is provided):** Fetches all users who do not currently have entries in the `UserTag` table.
    * **New User (if a user is provided):**  Creates entries only for the specified `user`.
    * **Tagged Fields:** Retrieves all tagged fields the `TaggedFieldModel`.

    For each user and tagged field combination, a new `UserTag` record is created with:

    * The corresponding user and tagged field.
    * Model and field names (both name and verbose name).
    * A unique slug based on the user's ID.
    * Default empty tags and a comment indicating that tags need to be added.

    The bulk creation of `UserTag` objects is performed within a database transaction to ensure data integrity.

    **Args:**
        user (User, optional): A specific user for whom to generate tags. If not provided, tags will be generated for all users missing from the `UserTag` table.

    **Raises:**
        ValidationError:
            * If a duplicate value is attempted to be created for a field with a unique constraint.
            * If invalid data types are provided for any of the `UserTag` fields.
            * If a general database error occurs during record creation.
    """
    stdout_with_optional_color(
        message="tag-me user tag generation tool is running", color_code=36
    )
    try:
        if user:
            users = [user]
            stdout_with_optional_color(
                message=f"A single user < {user.username} > requires updating",
                color_code=96,
            )
        else:
            # Get all unique users from UserTag
            existing_user_ids = UserTag.objects.values_list(
                "user_id", flat=True
            ).distinct()

            # Get all users who are NOT in existing_user_ids
            users = User.objects.exclude(id__in=existing_user_ids)

        user_count = len(users)
        match user_count:
            case 0:
                stdout_with_optional_color(
                    message="Nothing to do, all users are in the UserTag table!\nExiting user tag table generation tool...",
                    color_code=33,
                )
                return
            case 1:
                stdout_with_optional_color(
                    message=f"Generating UserTag rows for {user_count} user!",
                    color_code=36,
                )
            case _:
                stdout_with_optional_color(
                    message=f"Generating UserTag rows for {user_count} users!",
                    color_code=36,
                )
        tagged_fields = TaggedFieldModel.objects.all()
        user_tags = []
        for user in users:
            for field in tagged_fields:
                user_tags.append(
                    UserTag(
                        user=user,
                        tagged_field=field,
                        model_name=field.model_name,
                        model_verbose_name=field.model_verbose_name,
                        field_name=field.field_name,
                        field_verbose_name=field.field_verbose_name,
                        slug=TagBase.slugify(tag=str(user.id)),
                        tags=field.default_tags,
                        comment="Auto generated, please add tags and update/delete this comment",
                    )
                )
            with transaction.atomic():
                # Note: Bulk create UserTag objects, ignoring conflicts due to unique constraints.
                UserTag.objects.bulk_create(user_tags, ignore_conflicts=True)

        stdout_with_optional_color(
            message=f"    SUCCESS: Added {len(user_tags)} user tags rows in to the UserTag table for {len(users)} users!",
            color_code=92,
        )

    except IntegrityError as e:
        # Handle unique constraint violations
        if "duplicate key value violates unique constraint" in str(e):
            # Extract the conflicting value or field(s) if possible
            logger.exception(msg="IntegrityError")
            raise ValidationError(
                f"Duplicate value found for UserTag: {str(e)}"
            )
        else:
            raise

    except DataError as e:
        # Handle data type mismatches
        msg = f"Invalid data type provided for UserTag: {str(e)}"
        logger.exception(msg="DATA ERROR")
        raise ValidationError(msg)

    except DatabaseError as e:
        # Handle general database errors
        msg = f"Database error during UserTag creation: {str(e)}"
        logger.exception(msg="Database Error")
        raise ValidationError(msg)


def update_models_with_tagged_fields_table() -> None:
    """Updates the Tagged Field Models table for managing tagged fields.

    This function helps manage the models and fields in your Django project
    that use the custom 'TagMeCharField' for tagging.  Specifically, it does
    the following:

    1. Finds all the models that have fields using 'TagMeCharField'.
    2. Looks at each 'TagMeCharField' within these models.
    3. Adds or updates information about each tagged field in the
        'TaggedFieldModel' table.  This provides a centralized way to see which
        models and fields use tags.

    """
    for model in get_models_with_tagged_fields():
        content = ContentType.objects.get_for_model(
            model, for_concrete_model=True
        )
        model_name = content.model_class().__name__
        model_verbose_name = content.model_class()._meta.verbose_name
        for field in get_model_tagged_fields_field_and_verbose(
            model_name=model_name,
            return_field_objects_only=True,
        ):
            match field:
                case None:  # Ignore if specific field name is missing
                    # We test for None because the first tuple returned is None
                    pass
                case _:
                    (
                        obj,
                        created,
                    ) = TaggedFieldModel.objects.update_or_create(
                        content=content,
                        field_name=field.name,
                        field_verbose_name=field.verbose_name,
                        model_name=model_name,
                        model_verbose_name=model_verbose_name,
                        tag_type=field.tag_type,
                    )  # Add or update the database entry

                    match created:
                        case True:
                            logger.info(
                                f"\n-- Created {obj} : {field}"
                            )  # Log a new entry
                        case False:
                            logger.info(
                                f"\n-- Updated {obj} : {field}"
                            )  # Log an updated entry
        update_fields_that_should_be_synchronised()


def update_fields_that_should_be_synchronised():
    """
    Updates the synchronization configuration to include fields marked for tag synchronization.

    This function scans all models and checks for fields that have the 'synchronise'
    attribute set to True. If found, it updates the 'TagMeSynchronise' model (named 'default')
    to ensure that tags applied to those fields will be synchronised across relevant content types.
    """

    # Retrieve or create the default synchronization configuration
    sync, _ = TagMeSynchronise.objects.get_or_create(name="default")

    # Get a list of unique content IDs from existing TaggedFieldModel instances
    model_content_ids = (
        TaggedFieldModel.objects.all()
        .values_list("content_id", flat=True)
        .distinct()
    )

    # Retrieve ContentType models for further processing
    models_for_sync = ContentType.objects.filter(
        id__in=model_content_ids
    )  # Get models_for_sync from the list of content_id's

    # Flag to track changes to the sync config
    sync_updated: bool = False
    for model in models_for_sync:
        for field in model.model_class()._meta.fields:
            # Check if the field has the 'synchronise' attribute, and it's set
            # to True
            if hasattr(field, "synchronise") and field.synchronise:
                # Ensure the field is registered for synchronization
                if not sync.synchronise.get(field.name, False):
                    sync.synchronise[field.name] = []
                # Add the ContentType ID to the sync list for this field
                if model.id not in sync.synchronise[field.name]:
                    sync.synchronise[field.name].append(model.id)
                    sync_updated = True

    # Save the updated synchronization configuration if changes were made
    if sync_updated:
        sync.save()



"""example Core App Generic Views List."""

from core.views.generic.base import HtmxTemplateResponseMixin
from django.views.generic import ListView


class HtmxListView(HtmxTemplateResponseMixin, ListView):
    """Extends Django List view with HTMX functionality.

    Renders an htmx template or a generic template response.
    """



# Generated by Django 5.0.7 on 2024-08-03 22:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0002_alter_tagtestbase_tags"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tagtestbase",
            name="tags",
            field=models.CharField(
                blank=True,
                default="",
                help_text="This is the tag",
                max_length=255,
                null=True,
                verbose_name="tags",
            ),
        ),
    ]



"""blog Forms file."""

from django import forms

from tag_me.db.forms.mixins import TagMeModelFormMixin

from .models import Article


class ArticleCreateForm(TagMeModelFormMixin, forms.ModelForm):
    """Article tags creation form"""

    class Meta:
        model = Article
        fields = "__all__"
        exclude = ["tag", "user_tag2"]


class ArticleUpdateForm(TagMeModelFormMixin, forms.ModelForm):
    """Article tags creation form"""

    class Meta:
        model = Article
        fields = "__all__"



"""Management command to update the Tagged Field Models table."""

import logging

from django.core.management.base import BaseCommand, LabelCommand

from tag_me.utils.tag_mgmt_system import (
    generate_user_tag_table_records,
)


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def handle(self, *args, **options):
        """ """

        try:
            generate_user_tag_table_records()
        except Exception as e:
            logger.error(
                "Tags Table Update Error",
                exc_info=True,
            )
            self.stdout.write(
                self.style.ERROR(
                    "    ERROR: Tagged Models Table, and "
                    " Synchronised Fields not updated."
                )
            )



"""example project CustomUser Models."""

from django.apps import apps
from django.contrib import auth
from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.translation import gettext_lazy as _


class CustomUserManager(BaseUserManager):
    use_in_migrations = True

    def _create_user(
        self, username, email, password, **extra_fields
    ):  # pragma: no cover
        """
        Create and save a user with the given username, email, and password.
        """
        if not username:
            raise ValueError("A username must be supplied.")
        email = self.normalize_email(email)
        if not email:
            raise ValueError(_("An email address must be supplied."))
        if not password:
            raise ValueError(_("A password must be supplied."))
        # Lookup the real model class from the global app registry so this
        # manager method can be used in migrations. This is fine because
        # managers are by definition working on the real model.
        GlobalUserModel = apps.get_model(
            self.model._meta.app_label, self.model._meta.object_name
        )
        username = GlobalUserModel.normalize_username(username)
        user = self.model(username=username, email=email, **extra_fields)
        user.password = make_password(password)
        user.save(using=self._db)

        return user

    def create_user(
        self, username, email=None, password=None, **extra_fields
    ):  # pragma: no cover
        """
        Create and save a standard user with the supplied username, email, and password.
        """
        extra_fields.setdefault("is_active", True)
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return self._create_user(username, email, password, **extra_fields)

    def create_superuser(
        self, username, email=None, password=None, **extra_fields
    ):  # pragma: no cover
        """
        Create and save a super user with the supplied username, email, and password.
        """
        extra_fields.setdefault("is_active", True)
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)
        extra_fields.setdefault("user_type", "SUPERUSER")

        if extra_fields.get("is_staff") is not True:
            raise ValueError(_("Superuser must be staff."))
        if extra_fields.get("is_active") is not True:
            raise ValueError(_("Superuser must be an active user."))
        if extra_fields.get("is_superuser") is not True:
            raise ValueError(_("Superuser must be a superuser."))
        if extra_fields.get("user_type") != "SUPERUSER":
            raise ValueError(_("Superuser must be type SUPERUSER."))

        return self._create_user(username, email, password, **extra_fields)

    def with_perm(
        self,
        perm,
        is_active=True,
        include_superusers=True,
        backend=None,
        obj=None,
    ):  # pragma: no cover
        if backend is None:
            backends = auth._get_backends(return_tuples=True)
            if len(backends) == 1:
                backend, _ = backends[0]
            else:
                raise ValueError(
                    "You have multiple authentication backends configured and "
                    "therefore must provide the `backend` argument."
                )
        elif not isinstance(backend, str):
            raise TypeError(
                "backend must be a dotted import path string (got %r)."
                % backend
            )
        else:
            backend = auth.load_backend(backend)
        if hasattr(backend, "with_perm"):
            return backend.with_perm(
                perm,
                is_active=is_active,
                include_superusers=include_superusers,
                obj=obj,
            )
        return self.none()


class CustomUser(AbstractUser):
    """Extends the standard User model"""

    objects = CustomUserManager()

    class CustomUserType(models.TextChoices):
        """Custom user type choices"""

        # Change these to suit your needs before initial migration.
        FREE = "FREE", _("Free")
        LEVEL_1 = "LEVEL 1", _("Level 1")
        LEVEL_2 = "LEVEL 2", _("Level 2")
        STAFF = "STAFF", _("Staff")
        SUPERUSER = "SUPERUSER", _("Superuser")

    user_type = models.CharField(
        _("Type"),
        null=False,
        blank=False,
        max_length=50,
        choices=CustomUserType.choices,
        default=CustomUserType.FREE,
        help_text="Displays the users current user type.",
    )
    image_url = models.URLField(
        max_length=250,
        verbose_name=_(
            "Image URL",
        ),
        null=True,
        blank=True,
    )



"""Initialise example Core App Generic Views."""

from core.views.generic.base import HtmxTemplateView
from core.views.generic.detail import HtmxDetailView
from core.views.generic.edit import (
    HtmxCreateView,
    HtmxDeleteView,
    HtmxFormView,
    HtmxUpdateView,
)
from core.views.generic.list import HtmxListView

__all__ = [
    "HtmxCreateView",
    "HtmxDeleteView",
    "HtmxDetailView",
    "HtmxFormView",
    "HtmxListView",
    "HtmxTemplateView",
    "HtmxUpdateView",
    "HtmxGenericViewError",
]


class HtmxGenericViewError(Exception):
    """A problem in a generic view."""

    pass



"""django-tag-me Admin file."""

from django.contrib import admin

from tag_me.models import TaggedFieldModel, UserTag

# Register your tag_me models here.

admin.site.register(TaggedFieldModel)
admin.site.register(UserTag)



"""tag-me model field tests"""

from django.core import validators
from django.test import SimpleTestCase

from tag_me.db.forms.fields import TagMeCharField
from tag_me.utils.collections import FieldTagListFormatter


class TestTagMeCharFieldForm(SimpleTestCase):
    def test_empty_value(self):
        f = TagMeCharField()

        assert f.to_python("") == ""

    def test_with_value(self):
        f = TagMeCharField()

        assert f.to_python("a,b,c") == FieldTagListFormatter("a,b,c").toCSV(
            include_trailing_comma=True,
        )

    def test_validators(self):
        f = TagMeCharField(
            min_length=1234,
            max_length=5678,
        )

        assert any(
            x
            for x in f.validators
            if isinstance(x, validators.MaxLengthValidator)
        )

        assert any(
            x
            for x in f.validators
            if isinstance(x, validators.MinLengthValidator)
        )

        assert any(
            x
            for x in f.validators
            if isinstance(x, validators.ProhibitNullCharactersValidator)
        )



from django import template
from django.apps import apps
from django.utils.translation import pgettext_lazy as _

register = template.Library()


@register.filter
def get_app_verbose_name(content_type):
    try:
        return _(
            "App Verbose Name",
            apps.get_app_config(content_type.app_label).verbose_name,
        )
    except LookupError:
        return _(
            "App Label",
            content_type.app_label,
        )



from typing import Dict, Union

from django.forms import Field

from tag_me.db.forms.fields import TagMeCharField
from tag_me.models import TaggedFieldModel, UserTag


class TagMeModelFormMixin:
    """
    Mixin for Django ModelForms that enhances TagMeCharField fields.

    This mixin provides additional configuration and behavior for `TagMeCharField` fields within a ModelForm. It allows for customization based on the current user.
    """

    fields: Dict[str, Union[TagMeCharField, Field]]

    def __init__(self, *args, **kwargs):
        """
        Initializes the form, extracting the current user from keyword arguments.

        Args:
            *args: Variable length argument list passed to the parent ModelForm's __init__.
            **kwargs: Arbitrary keyword arguments passed to the parent ModelForm's __init__.
                 - user: (Optional) The current user object.
        """

        self.user = kwargs.pop("user", None)
        self.model_obj = kwargs.pop("model_obj", None)
        self.model_verbose_name = kwargs.pop("model_verbose_name", None)
        self.model_name = kwargs.pop("model_name", None)
        super().__init__(*args, **kwargs)  # Call the original form's __init__

        # Process fields
        for _, field in self.fields.items():
            # self.fields["field"].initial = obj._meta.get_field(field_name).verbose_name
            if isinstance(field, TagMeCharField):
                field.widget.attrs.update(
                    {
                        "css_class": "",
                        "user": self.user,
                    }
                )
                # self.fields["field"].initial = field



"""django-tag-me Views file."""

import base64
import binascii
import json
import logging

from django.core.exceptions import ObjectDoesNotExist
from django.http import JsonResponse
from django.urls import reverse_lazy
from django.views import View
from django.views.generic import TemplateView
from django.views.generic.edit import UpdateView
from django.views.generic.list import ListView

from tag_me.forms import (
    TaggedFieldEditForm,
    UserTagEditForm,
    UserTagListForm,
)
from tag_me.models import (
    TaggedFieldModel,
    TagMeSynchronise,
    UserTag,
)
from tag_me.utils.collections import FieldTagListFormatter

logger = logging.getLogger(__name__)


class TagManagementView(TemplateView):
    """Tag management dashboard"""

    template_name = "tag_me/mgmt/management.html"


class TaggedFieldEditView(UpdateView):
    """Edit the default tags available for users."""

    model = TaggedFieldModel
    form_class = TaggedFieldEditForm
    template_name = "tag_me/mgmt/edit_tagged_fields.html"
    success_url = reverse_lazy("tag_me:tag-mgmt")

    def form_valid(self, form):
        default_tags = FieldTagListFormatter(form.cleaned_data["default_tags"])
        form.cleaned_data["default_tags"] = default_tags.toCSV()

        return super().form_valid(form)


class TagFieldListView(ListView):
    """A list of all the tagged fields."""

    model = TaggedFieldModel
    template_name = "tag_me/mgmt/tagged_field_list.html"


class SynchronisedTagFieldListView(ListView):
    """A list of all the synchronised tagged fields."""

    model = TagMeSynchronise
    template_name = "tag_me/mgmt/sync_tagged_field_list.html"


class UserTagListView(ListView):
    """List user tags."""

    model = UserTag
    form_class = UserTagListForm
    template_name = "tag_me/user/list_user_tag.html"
    success_url = reverse_lazy("tag_me:tag-mgmt")

    def get_context_data(self, *, object_list=None, **kwargs):
        object_list = UserTag.objects.filter(user=self.kwargs["pk"])
        return super().get_context_data(object_list=object_list, **kwargs)


class MgmtUserTagListView(ListView):
    """List user tags."""

    model = UserTag
    form_class = UserTagListForm
    template_name = "tag_me/mgmt/list_user_tag.html"
    success_url = reverse_lazy("tag_me:tag-mgmt")


class UserTagEditView(UpdateView):
    """Update user tag."""

    model = UserTag
    form_class = UserTagEditForm
    template_name = "tag_me/user/edit_user_tag.html"
    success_url = reverse_lazy("tag_me:tag-mgmt")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["usertag"] = UserTag.objects.get(id=self.kwargs["pk"])
        return context

    def form_valid(self, form):
        usertag = UserTag.objects.get(id=self.kwargs["pk"])
        usertag.tags = form.cleaned_data["tags"]
        usertag.save()
        return super().form_valid(form)

    def form_invalid(self, form):
        return super().form_invalid(form)


class WidgetAddUserTagView(View):
    """Add tags from the TagMeCharfield widget."""

    def post(self, request, *args, **kwargs):
        """Handles POST requests to add tags to a UserTag instance."""

        encoded_data = request.POST.get("encoded_tag")

        if not encoded_data:
            return JsonResponse(
                {"error": "Invalid or corrupted tag data"}, status=400
            )

        try:
            data = json.loads(
                base64.urlsafe_b64decode(encoded_data).decode("utf-8")
            )
            try:
                user_tag = UserTag.objects.get(id=self.kwargs["pk"])
            except ObjectDoesNotExist:
                return JsonResponse({"error": "UserTag not found"}, status=404)
            all_tags = FieldTagListFormatter(user_tag.tags)
            all_tags.add_tags(data)
            user_tag.tags = all_tags.toCSV()
            user_tag.save()

            return JsonResponse(
                {
                    "success": True,
                    "tags": all_tags.toList(),
                }
            )
        except (json.JSONDecodeError, binascii.Error):
            logger.exception("Error decoding tag data.")
            return JsonResponse(
                {"error": "Invalid or corrupted data"}, status=400
            )



"""Django staging settings for example project."""

import environ
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .base import *  # noqa: F405 F401 F403

env = environ.Env(
    # set casting, default value
    DEBUG=(bool, False)
)
environ.Env.read_env(os.path.join(BASE_DIR, ".env/.staging"))  # noqa: F405

logger = logging.getLogger(__name__)


ALLOWED_HOSTS = env.list(
    "STAGING_ALLOWED_HOSTS",
    default=[""],
)

DEBUG = env("STAGING_DJANGO_DEBUG", default=False)


DJANGO_LOGGING_MAIL_ADMINS = env(
    "STAGING_DJANGO_LOGGING_MAIL_ADMINS",
    default="ERROR",
)

DJANGO_LOGGING_LEVEL = env(
    "STAGING_DJANGO_LOGGING_LEVEL",
    default="INFO",
)

DJANGO_LOG_FILE = env(
    "STAGING_DJANGO_LOG_FILE",
    default="logging/rotating.log",
)

DJANGO_SETTINGS_MODULE = env(
    "STAGING_DJANGO_SETTINGS_MODULE",
    default="config.settings.staging",
)


DATABASES = {
    "default": env.db(
        "STAGING_DATABASE_URL",
    ),
}

EMAIL_BACKEND = env(
    "STAGING_EMAIL_BACKEND",
    default="django.core.mail.backends.console.EmailBackend",
)


# Default empty string required for Docker build to succeed.
EMAIL_HOST = env(
    "STAGING_EMAIL_HOST",
    default="",
)


EMAIL_HOST_PASSWORD = env(
    "STAGING_EMAIL_HOST_PASSWORD",
    default="",
)

EMAIL_HOST_USER = env(
    "STAGING_EMAIL_HOST_USER",
    default="",
)
EMAIL_PORT = env(
    "STAGING_EMAIL_PORT",
    default="",
)
EMAIL_USE_TLS = env(
    "STAGING_EMAIL_USE_TLS",
    default="",
)

INTERNAL_IPS = env.list(
    "STAGING_INTERNAL_IPS",
    default=[""],
)

# Override the default logger level to the django environment
# log Level settings
LOGGING["loggers"][""]["level"] = DJANGO_LOGGING_LEVEL  # noqa: F405
LOGGING["handlers"]["stdout"]["level"] = DJANGO_LOGGING_LEVEL  # noqa: F405
LOGGING["handlers"]["rotated_logs"][
    "level"
] = DJANGO_LOGGING_LEVEL  # noqa: F405
LOGGING["handlers"]["rotated_logs"]["filename"] = DJANGO_LOG_FILE  # noqa: F405


SECRET_KEY = env(
    "STAGING_DJANGO_SECRET_KEY",
    default="!!!INSECURE_STAGING_SECRET!!!",
)


# Check it is safe to run in staging.
assert (  # nosec
    env("STAGING_DJANGO_SECRET_KEY") != "!!!INSECURE_STAGING_SECRET!!!"
), "The DJANGO_SECRET_KEY must be set for staging."


# `USE_STATIC` options are `local` or `S3`
USE_STATIC = env("STAGING_USE_STATIC", default="Local")


try:
    # Digital Ocean S3 Storage Configuration
    if USE_STATIC == "S3":
        AWS_ACCESS_KEY_ID = env(
            "STAGING_AWS_ACCESS_KEY_ID",
            default="STAGING_AWS_KEY_NOT_SET",
        )
        AWS_SECRET_ACCESS_KEY = env(
            "STAGING_AWS_SECRET_ACCESS_KEY",
            default="STAGING_AWS_SECRET_NOT_SET",
        )

        AWS_S3_REGION_NAME = env(
            "STAGING_AWS_S3_REGION_NAME",
            default="syd1",
        )
        AWS_S3_ENDPOINT_URL = env(
            "STAGING_AWS_S3_ENDPOINT_URL",
            default=f"https://{AWS_S3_REGION_NAME}.digitaloceanspaces.com",
        )
        AWS_STORAGE_BUCKET_NAME = env(
            "STAGING_AWS_STORAGE_BUCKET_NAME",
            default="tb-s3",
        )
        AWS_LOCATION = env(
            "STAGING_AWS_LOCATION",
            default=f"https://{AWS_STORAGE_BUCKET_NAME}.{AWS_S3_REGION_NAME}.digitaloceanspaces.com",
        )
        # This isnt working when using an ENV VAR
        # AWS_S3_OBJECT_PARAMETERS = env(
        #     "STAGING_AWS_S3_OBJECT_PARAMETERS",
        #     {"CacheControl": "max-age=1"},
        # )
        AWS_S3_OBJECT_PARAMETERS = {"CacheControl": "max-age=1"}

        STORAGES = {
            "default": {
                "BACKEND": env(
                    "DEFAULT_FILE_STORAGE",
                    default="core.storage.backends.MediaRootS3Boto3Storage",
                )
            },
            "staticfiles": {
                "BACKEND": env(
                    "STATICFILES_STORAGE",
                    default="core.storage.backends.StaticRootS3Boto3Storage",
                )
            },
        }

        # Static url must end with default STATIC_URL from env var or base.py added to
        # S3 storage location.

        STATIC_URL = env("STAGING_STATIC_URL", default="static-st/")

        # Media url must end with default MEDIA_URL from env var or base.py added to
        # S3 storage location.
        MEDIA_URL = env("STAGING_MEDIA_URL", default="media-st/")

        # Set the url for the css file
        STAGING_DJANGO_TEMPLATES_CSS = f"{STATIC_URL}css/styles.css"

    elif USE_STATIC != "S3":
        raise ImproperlyConfigured(
            "Staging environment USE_STATIC must be S3, it is configured to %s."
            % (USE_STATIC),
        )

except ImproperlyConfigured:
    logger.critical(
        "Staging environment USE_STATIC must be S3, it is configured to %s."
        % (USE_STATIC),
    )



"""tag-me app custom form widget."""

import json

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe

from tag_me.models import UserTag

User = get_user_model()


class TagMeSelectMultipleWidget(forms.SelectMultiple):
    allow_multiple_selected = True

    # @override
    def render(self, name, value, attrs=None, renderer=None) -> str:
        """Renders a multiple select HTML element with dynamically generated choices.  # noqa: E501

        A custom Django form widget that provides user-specific options
        tailored to a particular model field. It's designed to be flexible and
        works by fetching relevant tag choices on the fly.

        Args:
            :param name: The name attribute to use for the generated
                         <select> element.
            :param value:  The currently selected value or values for the
                           field.  This can be a single value or potentially a
                           list/iterable of values for multiple selection.
            :param attrs: (dict, optional) A dictionary of additional
                          attributes to include in the rendered <select> tag
                          (e.g., 'id', 'class')
            :param renderer:  (Django Renderer, optional) An advanced option
                                            to override the rendering engine.
                                            Most users can ignore this.

        Returns:
            str:  Mark safe HTML output representing the fully formed <select>
                  element with its <option> tags populated from your dynamic
                  choices.
        """
        # Important: 'attrs' is modified in place by removing some entries
        # The 'attrs' removed are for filtering choices and not required
        # elsewhere.
        css_class = self.attrs.get("css_class", None)
        _field_verbose_name = self.attrs.pop("field_verbose_name", None)
        _allow_multiple_select = self.attrs.pop("allow_multiple_select", True)
        _auto_select_new_tags = self.attrs.pop("auto_select_new_tags", True)
        _tag_choices = self.attrs.pop("tag_choices", None)
        _tagged_field = self.attrs.pop("tagged_field", None)
        user = self.attrs.pop("user", None)

        if "display_number_selected" not in self.attrs:
            _display_number_selected = settings.DJ_TAG_ME_MAX_NUMBER_DISPLAYED

        # Get the template theme
        if "theme" not in self.attrs:
            self.template_name = settings.DJ_TAG_ME_THEMES["default"]
        else:
            self.template_name = settings.DJ_TAG_ME_THEMES[self.attrs["theme"]]

        _add_tag_url = ""
        _permitted_to_add_tags = True

        # Call the parent class render (essential for Widget functionality)
        super().render(name, value, attrs, renderer)

        if _tag_choices:
            # Here we are using the choices set in the model charfield.
            self.choices = _tag_choices
            _permitted_to_add_tags = False
        else:
            # Dynamically fetch user and field specific choices as a list.
            user_tags = UserTag.objects.filter(
                user=user,
                tagged_field=_tagged_field,
            ).first()

            if user_tags.tags:
                self.choices = [tag.strip() for tag in user_tags.tags.split(",")]
            else:
                self.choices = []
            _add_tag_url = reverse("tag_me:add-tag", args=[user_tags.id])

        values: list = []
        match value:
            case str():
                for val in value.rstrip(",").split(","):
                    values.append(val.strip())
        context = {
            "add_tag_url": _add_tag_url,
            "allow_multiple_select": json.dumps(_allow_multiple_select),
            "auto_select_new_tags": json.dumps(_auto_select_new_tags),
            "choices": self.choices,
            "display_number_selected": _display_number_selected,
            "name": name,
            "permitted_to_add_tags": json.dumps(_permitted_to_add_tags),
            "verbose_name": _field_verbose_name,
            "values": values,
            # "options": json.dumps(options),
        }

        return mark_safe(render_to_string(self.template_name, context))



# Generated by Django 5.0.6 on 2024-07-23 07:39

import django.db.models.deletion
import tag_me.db.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Post",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    tag_me.db.models.fields.TagMeCharField(
                        blank=True, max_length=255, null=True
                    ),
                ),
                ("body", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="TagTestBase",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "tags",
                    models.CharField(
                        help_text="This is the tag",
                        max_length=50,
                        verbose_name="tags",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        allow_unicode=True,
                        editable=False,
                        max_length=100,
                        unique=True,
                        verbose_name="slug",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tag ABC",
                "verbose_name_plural": "Tags ABC",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="TaggedFieldTestModel",
            fields=[
                (
                    "tagtestbase_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.tagtestbase",
                    ),
                ),
                (
                    "tagged_field_1",
                    tag_me.db.models.fields.TagMeCharField(
                        blank=True, max_length=255, null=True
                    ),
                ),
                (
                    "tagged_field_2",
                    tag_me.db.models.fields.TagMeCharField(
                        blank=True, max_length=255, null=True
                    ),
                ),
            ],
            options={
                "verbose_name": "Tagged Field Test Model",
                "verbose_name_plural": "Tagged Field Test Models",
            },
            bases=("tests.tagtestbase",),
        ),
    ]



"""Management command to update the Tagged Field Models table."""

import logging

from django.conf import settings
from django.core.management.base import BaseCommand, LabelCommand

from tag_me.utils.tag_mgmt_system import (
    update_models_with_tagged_fields_table,
    generate_user_tag_table_records,
)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def handle(self, *args, **options):
        """
        Handles the execution of tag-related updates and synchronization checks.

        This method performs the following:
            1. Logs the start of the update process.
            2. Updates the 'TaggedFieldModel' table for tag management.
            3. Checks and updates field synchronization configurations.
            4. Reports any errors encountered during the update process.

        :param args: Additional positional arguments.
        :param options: Additional keyword arguments.
        """
        logger.info("Updating Tagged Models Table, and Synchronised Fields.")

        try:
            self.stdout.write("    Updating Tagged Models Table.")

            update_models_with_tagged_fields_table()

            generate_user_tag_table_records()

            self.stdout.write(
                self.style.SUCCESS(
                    "    SUCCESS: Tagged Models Table,"
                    " and Synchronised Fields updated."
                )
            )
            from tag_me.models import TagMeSynchronise

            sync, _ = TagMeSynchronise.objects.get_or_create(name="default")
            sync.check_field_sync_list_lengths()

        except Exception as e:
            logger.error(
                "Tags Table Update Error",
                exc_info=True,
            )
            self.stdout.write(
                self.style.ERROR(
                    "    ERROR: Tagged Models Table, and "
                    " Synchronised Fields not updated."
                )
            )

        # finally:
        #     print("TAGS.PY OK, STARTING FINALLY")
        #     # Ensure synchronization configuration check is performed and logged.
        #     from tag_me.models import TagMeSynchronise
        #
        #     sync, _ = TagMeSynchronise.objects.get_or_create(name="default")
        #     sync.check_field_sync_list_lengths()



"""Custom migrate command"""

import logging
import subprocess
from io import StringIO

from django.conf import settings
from django.core.management import call_command
from django.core.management.commands.migrate import (
    Command as CoreMigrateCommand,
)

logger = logging.getLogger(__name__)


class Command(CoreMigrateCommand):
    """Override `migrate.py` to run the `tag_me` tables update command.

    Call the tags management command after migrate.
    tag_me management command `tags.py` checks all models and adds any with
    tag fields to `TaggedFieldModel` table.
    """

    def handle(self, *args, **options):
        super().handle(*args, **options)
        out = StringIO()
        call_command(
            "tags",
            stdout=out,
        )
        return out.getvalue()

        # def handle(self, *args, **options):
        if settings.DJ_TAG_ME_USE_CUSTOM_MIGRATE:
            super().handle(*args, **options)

    #
    #     err = None
    #     out = None
    #
    #     try:
    #         with subprocess.Popen(  # nosec
    #             args=["./manage.py", "tags"],
    #             stdout=subprocess.PIPE,
    #             stderr=subprocess.PIPE,
    #             bufsize=0,
    #             text=True,
    #             encoding="utf-8",
    #         ) as proc:
    #             out, err = proc.communicate(timeout=10)
    #
    #             logger.info(f"Process Successful: {args}\n{out}")
    #
    #     except subprocess.TimeoutExpired:
    #         logger.error(f"{args} error: {err}")
    #         logger.error("Exiting migrate!")
    #         raise SystemExit(1)
    #
    #     except subprocess.CalledProcessError as e:
    #         logger.error(
    #             f"{args} failed with return code {proc.returncode}.\nError {e}"
    #         )
    #         logger.error("Exiting nuke_db_sqlite")
    #         raise SystemExit(1)
    #
    #     except OSError as e:
    #         print(f"ERROR IS {e}")
    #         logger.error(f"{args} failed with return code.\nError{e}")
    #         logger.error("Exiting nuke_db_sqlite")
    #         raise SystemExit(1)
    #
    #     except ValueError as e:
    #         logger.error(f"{args} failed with return code {proc.returncode}.\nError{e}")
    #         logger.error("Exiting nuke_db_sqlite")
    #         raise SystemExit(1)
    #
    #     # return out
    #
    #     # logger.error("######### TAGS MIGRATE")
    #     # if settings.DJ_TAG_ME_USE_CUSTOM_MIGRATE:
    #     #     super().handle(*args, **options)
    #     # out = StringIO()
    #     # call_command(
    #     #     "tags",
    #     #     stdout=out,
    #     # )
    #     #
    #     # return out.getvalue()



"""django-tag-me view mixins"""

import logging

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.views.generic.edit import FormMixin

from tag_me.db.forms.mixins import TagMeModelFormMixin

logger = logging.getLogger(__name__)


class TagMeViewMixin(FormMixin):
    """
    Mixin for Django CBVs to handle forms with tagged fields.
    Passes the model's verbose name and current user to the TagMeModelForm.
    """

    request: HttpRequest

    def get_form(self, form_class=None):
        """
        Overrides the get_form method to pass the user and model_verbose_name.
        """
        if form_class is None:
            form_class = self.get_form_class()

        if not issubclass(form_class, TagMeModelFormMixin):
            msg = f"The form {form_class} used with TagMeViewMixin must inherit from TagMeModelFormMixin."
            logger.exception(msg)
            raise ImproperlyConfigured(msg)

        form_kwargs = self.get_form_kwargs()

        # Check if user is already in form_kwargs (potentially from another mixin or view)
        if "user" not in form_kwargs:
            form_kwargs["user"] = self.request.user
        form_kwargs["model_verbose_name"] = (self.model._meta.verbose_name,)
        form_kwargs["model_obj"] = self.model
        return form_class(
            **form_kwargs,
        )

    def get_initial(self):
        """Return the initial data to use for forms on this view."""
        initial = super().get_initial()
        initial["user"] = self.request.user
        initial["content_type"] = ContentType.objects.get_for_model(
            self.model, for_concrete_model=True
        )
        initial["model_verbose_name"] = self.model.__dict__["_meta"].__dict__[
            "verbose_name"
        ]
        return initial



"""django-tag-me blog models"""

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import pgettext_lazy as _
from users.models import CustomUser

from tag_me.db.models.fields import TagMeCharField

User = get_user_model()


class Author(models.Model):
    """Author model"""

    class Meta:
        verbose_name = _(
            "Verbose Name",
            "Author",
        )
        verbose_name_plural = _(
            "Verbose Plural",
            "Authors",
        )

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )
    pen_name = models.CharField(
        max_length=255,
        verbose_name="Author",
    )

    bio = models.TextField(
        blank=True,
        null=True,
        verbose_name="Biography",
    )

    def __str__(self):
        return self.pen_name


class Article(models.Model):
    """Article model"""

    class Meta:
        verbose_name = "Blog Article"
        verbose_name_plural = "Blog Articles"

    class ArticleTags(models.TextChoices):
        """An example using choices to constrain tag list"""

        TWOMINREAD = (
            "2 Min Read",
            _(
                "Article Tag",
                "2 Min Read",
            ),
        )
        FIVEMINREAD = (
            "5 Min Read",
            _(
                "Article Tag",
                "5 Min Read",
            ),
        )

    author = models.ForeignKey(
        "Author",
        on_delete=models.CASCADE,
        verbose_name="Article Author",
    )

    article = models.TextField(
        verbose_name="Article",
    )
    tag = TagMeCharField(
        max_length=255,
        null=True,
        blank=True,
        choices=ArticleTags.choices,
        verbose_name="Article Tag",
        help_text="A tag for the Article",
    )

    user_tag = TagMeCharField(
        max_length=255,
        null=True,
        blank=True,
        verbose_name="Article User Tag",
        help_text="A tag for the Article",
    )

    user_tag2 = TagMeCharField(
        max_length=255,
        null=True,
        blank=True,
        verbose_name="Article User Tag 2",
        help_text="A tag2 for the Article",
    )



"""example URL Configuration.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

# from allauth import account
from allauth import account
from django.conf import settings
from django.contrib import admin
from django.urls import include, path

urlpatterns = [
    path("admin/", admin.site.urls),
    path("accounts/", include("allauth.urls")),
    path("", account.views.LoginView.as_view(), name="login"),
    path("tag_me/", include("tag_me.urls", namespace="tag_me")),
    path("blog/", include("blog.urls", namespace="blog")),
]


# If we are in DEBUG mode add debug-toolbar to url_patterns.
if settings.DEBUG:
    import debug_toolbar

    urlpatterns += [
        path("__debug__/", include(debug_toolbar.urls)),
        path("__reload__/", include("django_browser_reload.urls")),
    ]



"""A Django command line tool to delete migration files and dbsqlite in local development."""

import os
import subprocess  # nosec
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand

from tag_me.management.commands import migrate as tag_me_migrate


class Command(BaseCommand):
    """
    Delete app migrations, dbsqlite then makemigrations, migrate and load fixtures.

    .. warning::

        This only deletes the db and migration files.

        Migrations and fixture loading is not implemtented.

    """

    def __init__(self, request=None, *args, **kwargs):
        """Initialise the Basecommand with some local variables.

        Paramaters
        ----------

        :param PROJECT_APPS: A list of only the projects applications.
        :param DB_PATH: A posix path to the local dbsqlite3 db.
        :param PROJECT_ROOT: A posix path to the projects root folder.

        """

        # TODO Fix this...Make sure any changes to migrations are
        # reflected in the testing app.  Makemigrations using this doesnt work.
        # testing_core.migrations is excluded from file deletions in the mean time.
        self.PROJECT_APPS = settings.PROJECT_APPS + [
            "testing_core",
        ]  # noqa: F405

        self.DB_PATH = Path(settings.BASE_DIR / "db.sqlite3")  # noqa: F405
        self.PROJECT_ROOT = Path(settings.BASE_DIR)  # noqa: F405

        self.request = request

        super().__init__(*args, **kwargs)

    def _check_in_debug(self) -> bool:
        """Check that we are in DEBUG and wont be breaking Production DB"""
        if settings.DEBUG:
            return True
        else:
            self.stdout.write(self.style.ERROR("Exiting, not in DEBUG!"))
            return False

    def _delete_dbsqlite3(self) -> bool:
        """If dbsqlite exist, delete and return true, else exit."""

        if Path(self.DB_PATH).is_file():
            # If the database file exists alert user in the console of  the
            # actions taken or ....>
            try:
                self.stdout.write(
                    self.style.SUCCESS(
                        f"DATABASE PATH EXISTS: {str(self.DB_PATH)}"
                    )
                )
                self.stdout.write(
                    self.style.WARNING(
                        f"\tDELETING DATABASE: {str(self.DB_PATH)}"
                    )
                )

                # Delete DB
                self.DB_PATH.unlink()

                if not Path(self.DB_PATH).is_file():
                    self.stdout.write(
                        self.style.SUCCESS(
                            f"\tSUCESSFULLY DELETED {str(self.DB_PATH)}"
                        )
                    )

                return True

            except OSError as e:
                # <... Alert user in console if db failed to delete.

                self.stdout.write(
                    self.style.ERROR(
                        f"\nDELETION ERROR : {str(self.DB_PATH)}::{e.strerror}"
                    )
                )

                return False
        else:
            self.stdout.write(
                self.style.ERROR(
                    f"\n\nERROR: db.sqlite3 {str(self.DB_PATH)} Does not Exist"
                )
            )

            return False

    def _delete_migration_files(
        self, delete_app: str = None
    ) -> None:  # noqa: C901
        """Find, list and delete the required migration files.

        Paramaters
        ----------

        :param delete_app:
            User supplied string of the ``app name`` for migration deletions.


        Private Params
        --------------

        :param _delete_migrations:  :py:class: ``Dict``
                key: App name as string.
                value:  _migration_files_for_deletion as List.

        :param _deletion_errors:  :py:class: ``List``
                A list of ``strings`` of the migration file with ``OS.errors``.

        .. warning::

            This feature is currently incomplete.
            Only full migration deletions has been implemented.


        """
        _delete_migrations: dict = {}
        _deletion_errors: list = []

        if not delete_app:
            """Delete all migration files."""

            for app_name in self.PROJECT_APPS:
                # Check that the path is a valid directory.
                if Path(
                    self.PROJECT_ROOT
                    / app_name.replace(".", "/")
                    / "migrations"
                ).is_dir():
                    self.stdout.write(f"MARKED FOR DELETION::{app_name}")

                    _migration_files_for_deletion: list = []
                    # Get a list of migration files in the app and add to
                    # a deletion dict as a file path.
                    # TODO:
                    for file_path in Path(
                        self.PROJECT_ROOT
                        / app_name.replace(".", "/")
                        / "migrations"
                    ).iterdir():
                        # Dont delete __init__ , testing_core and check the
                        # path is a file first.
                        if (
                            "__init__" not in str(file_path)
                            and "testing_core" not in str(file_path)
                            and file_path.is_file()
                        ):
                            self.stdout.write(
                                self.style.WARNING(f"\t {file_path}")
                            )

                            _migration_files_for_deletion.append(file_path)

                    _delete_migrations[app_name] = (
                        _migration_files_for_deletion
                    )

            for app_name, migration_file_path in _delete_migrations.items():
                if migration_file_path:
                    # Alert user in console that the app migrations deletion is
                    # in progress for.
                    self.stdout.write(
                        self.style.NOTICE(
                            f"Deleting Migrations for App:{app_name}"
                        )
                    )

                    for file_path in migration_file_path:
                        try:
                            file_path.unlink()
                            # If file deleted ok,  print success message to console.
                            if not file_path.is_file():
                                self.stdout.write(
                                    self.style.SUCCESS(
                                        f"\t\t  Delete Successful: \t{str(file_path)}"
                                    )
                                )

                        except OSError as e:
                            _deletion_errors.append(
                                f"{file_path}::ERROR::{e.strerror}"
                            )

                else:
                    # Alert user in the console if no migration files exist.
                    self.stdout.write(
                        self.style.WARNING(
                            f"No migration files to delete\tApp:{app_name}"
                        )
                    )

            if _deletion_errors:
                # Alert user in console of file deletion errors.
                self.stdout.write(
                    self.style.WARNING("Migration Deletion Errors\n")
                )

                for file_error in _deletion_errors:
                    self.stdout.write(self.style.ERROR(f"\t {file_error}"))

        else:
            self.stdout.write(
                self.style.ERROR(
                    f"SINGLE APP:{delete_app} DELETION IS NOT IMPLEMENTED \n Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

            # TODO Complete this functionality.
            # Delete only user specified app migration files.
            # Check if the user supplied app exist and if not ...>
            # if delete_app in self.PROJECT_APPS:
            #     self.stdout.write(
            #         self.style.WARNING(f"Delete migrations for App:{delete_app}")
            #     )
            # else:
            #     # <... Alert user in console that app name does not exist
            #     self.stdout.write(
            #         self.style.ERROR(f"Error: {delete_app} does not exist.")
            #     )

    # +++++++++++++ manage.py methods +++++++++++++

    def _run_manage_py_with_command(
        self, *args, supress_exception=False, cwd=None
    ):
        """Helper to set up Django manage.py in a subprocess."""
        cur_dir = os.getcwd()

        try:
            if cwd:
                os.chdir(cwd)

            with subprocess.Popen(  # nosec
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ) as proc:
                out, err = proc.communicate()
                out = out.decode("utf-8")
                err = err.decode("utf-8")
                if err and not supress_exception:
                    raise Exception(err)
                if err and supress_exception:
                    return out

                return out

        finally:
            os.chdir(cur_dir)

    def _makemigrations(self) -> bool:
        """Instantiate a subprocess and makemigrations."""

        try:
            self._run_manage_py_with_command(
                "./manage.py",
                "makemigrations",
                supress_exception=False,
                cwd=self.PROJECT_ROOT,
            )

            self.stdout.write(self.style.SUCCESS("SUCCESS: makemigrations"))

            return True

        except OSError as e:
            self.stdout.write(
                self.style.ERROR(
                    f"makemigrations::ERROR::{e.strerror}::Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

    def _migrate(self) -> bool:
        """Instantiate a subprocess and migrate."""

        try:
            self._run_manage_py_with_command(
                "./manage.py",
                "migrate",
                supress_exception=False,
                cwd=self.PROJECT_ROOT,
            )

            self.stdout.write(self.style.SUCCESS("SUCCESS: migrate"))

            return True

        except OSError as e:
            self.stdout.write(
                self.style.ERROR(
                    f"migrate::ERROR::{e.strerror}::Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

    def _createsuperuser_ryan(self) -> bool:
        """Instantiate a subprocess and createsuperuser_ryan."""

        try:
            self._run_manage_py_with_command(
                "./manage.py",
                "createsuperuser_for_testing",
                "--username=user1",
                "--password=user1",
                "--email=user1@user1.com",
                supress_exception=False,
                cwd=self.PROJECT_ROOT,
            )

            self.stdout.write(
                self.style.SUCCESS("SUCCESS: createsuperuser::ryan")
            )

            return True

        except OSError as e:
            self.stdout.write(
                self.style.ERROR(
                    f"createsuperuser-ryan::ERROR::{e.strerror}::Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

    def _createsuperuser_mark(self) -> bool:
        """Instantiate a subprocess and createsuperuser_mark."""

        try:
            self._run_manage_py_with_command(
                "./manage.py",
                "createsuperuser_for_testing",
                "--username=user2",
                "--password=user2",
                "--email=user2@user2.com",
                supress_exception=False,
                cwd=self.PROJECT_ROOT,
            )

            self.stdout.write(
                self.style.SUCCESS("SUCCESS: createsuperuser::mark")
            )

            return True

        except OSError as e:
            self.stdout.write(
                self.style.ERROR(
                    f"createsuperuser-mark::ERROR::{e.strerror}::Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

    def _loadfixtures(self):
        fixtures_path = Path(settings.BASE_DIR / "core/management/fixtures")

        fixtures_list = list(x for x in fixtures_path.iterdir())
        fixtures_list.sort()

        print(f"\nFIXTURES {fixtures_list}")

        fixtures_list_print: str = ""

        for fix in fixtures_list:
            fixtures_list_print += str(fix) + "\n\t"

        self.stdout.write(
            self.style.WARNING(
                f"Loading {len(fixtures_list)} fixtures\n\t{fixtures_list_print}"
            )
        )

        try:
            for fixture in fixtures_list:
                self._run_manage_py_with_command(
                    "./manage.py",
                    "loaddata",
                    fixture,
                    supress_exception=False,
                    cwd=self.PROJECT_ROOT,
                )

                self.stdout.write(
                    self.style.SUCCESS(f"SUCCESS: Loading {fixture}")
                )

                # return True

        except OSError as e:
            self.stdout.write(
                self.style.ERROR(
                    f"Loading fixture {fixture}::ERROR::{e.strerror}::Exiting nuke_db_sqlite"
                )
            )

            raise SystemExit(1)

    # +++++++++++++ BASECOMMAND Overrides +++++++++++++

    def add_arguments(self, parser):
        parser.add_argument(
            "-D",
            "--delete_dbsqlite",
            action="store_true",
            default=True,
            help="Delete Django Project dbsqlite3",
        )

        parser.add_argument(
            "-A",
            "--app_name",
            # required=True
            action="store_true",
            help="Provide the Django App name for migrations deletion",
            default=None,
        )

        parser.add_argument(
            "-F",
            "--fixtures",
            # required=True
            action="store_true",
            help="Delete all migrations, migrate and load fixtures",
            default=None,
        )

    def handle(self, *args, **options):
        if self._check_in_debug():
            if options["fixtures"]:
                self._delete_dbsqlite3()
                self._delete_migration_files()
                self._makemigrations()
                self._migrate()
                self._loadfixtures()

            if options["delete_dbsqlite"]:
                if (
                    not input(
                        "You are about to delete all migrations and your local dbsqlite3 DB, proceed [y,n]"
                    )
                    == "y"
                ):
                    self.stdout.write(
                        self.style.ERROR(
                            "User aborted::Exiting nuke_db_sqlite"
                        )
                    )

                    raise SystemExit(2)

                if self._delete_dbsqlite3():
                    # If the database has been deleted successfully delete migrations.
                    self._delete_migration_files()
                    self._makemigrations()
                    self._migrate()

                    valid_choices: list = ["", "1", "2", "3"]

                    choice = input(
                        "Select a value default [1].\nEnter [1] Load fixtures.\nEnter [2] Create two superusers.\nEnter [3] Exit with an empty database."
                    )

                    counter: int = 1
                    while choice not in valid_choices:
                        if counter == 3:
                            self.stdout.write(
                                self.style.ERROR(
                                    "Exiting nuke_db, to many failed selection attempts.\n\tMigrations are complete and you have an empty database."
                                )
                            )

                            raise SystemExit(1)

                        choice = input(
                            "That option is not available.\nPlease select from options below default[1].\nEnter [1] Load fixtures.\nEnter [2] Create two superusers.\nEnter [3] Exit with an empty database."
                        )
                        counter += 1

                    match choice:
                        case "":
                            self._loadfixtures()

                            self.stdout.write(
                                self.style.SUCCESS(
                                    "Exiting nuke_db.\n\tMigrations are complete.\n\tYour fixtures have been loaded."
                                )
                            )

                            raise SystemExit(0)

                        case "1":
                            self._loadfixtures()

                            self.stdout.write(
                                self.style.SUCCESS(
                                    "Exiting nuke_db.\n\tMigrations are complete.\n\tYour fixtures have been loaded."
                                )
                            )

                            raise SystemExit(0)

                        case "2":
                            self._createsuperuser_ryan()
                            self._createsuperuser_mark()

                            self.stdout.write(
                                self.style.SUCCESS(
                                    "Exiting nuke_db.\n\tMigrations are complete.\n\tYou have two superusers in your database."
                                )
                            )

                            raise SystemExit(0)

                        case "3":
                            self.stdout.write(
                                self.style.SUCCESS(
                                    "Exiting nuke_db.\n\tMigrations are complete and you have an empty database."
                                )
                            )

                            raise SystemExit(0)

                else:
                    self.stdout.write(
                        self.style.ERROR("ERROR: Exiting nuke_db_sqlite")
                    )
                    # https://adamj.eu/tech/2021/10/10/the-many-ways-to-exit-in-python/
                    raise SystemExit(1)

            else:
                self.stdout.write(
                    self.style.ERROR(
                        "Deleting only Migration files, DB will not be deleted."
                    )
                )

                self._delete_migration_files()
        else:
            self.stdout.write(
                self.style.ERROR(
                    "ERROR: Not in DEBUG mode::Exiting nuke_db_sqlite"
                )
            )
            # https://adamj.eu/tech/2021/10/10/the-many-ways-to-exit-in-python/
            # https://www.youtube.com/watch?v=ZbeSPc5wL0g explains why to use this approach.
            raise SystemExit(1)



from io import StringIO

from django.core.management import call_command
from hypothesis.extra.django import TestCase


class TestCustomManagementCommands(TestCase):
    def test_custom_migrate_command_ok(self):
        self.out_migration = StringIO()
        call_command(
            "migrate",
            stdout=self.out_migration,
        )

        assert "Updating Tagged Models Table." in self.out_migration.getvalue()
        assert (
            "    SUCCESS: Tagged Models Table,"
            " and Synchronised Fields updated."
            in self.out_migration.getvalue()
        )



# Generated by Django 5.0.6 on 2024-08-03 23:04

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="SystemTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "tags",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="This is the tag",
                        max_length=255,
                        null=True,
                        verbose_name="tags",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        allow_unicode=True,
                        editable=False,
                        max_length=100,
                        unique=True,
                        verbose_name="slug",
                    ),
                ),
            ],
            options={
                "verbose_name": "System Tag",
                "verbose_name_plural": "System Tags",
            },
        ),
        migrations.CreateModel(
            name="TagMeSynchronise",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default="default", max_length=255)),
                (
                    "synchronise",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
            ],
            options={
                "verbose_name": "Tags Synchronised",
                "verbose_name_plural": "Tags Synchronised",
            },
        ),
        migrations.CreateModel(
            name="UserTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "tags",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="This is the tag",
                        max_length=255,
                        null=True,
                        verbose_name="tags",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        allow_unicode=True,
                        editable=False,
                        max_length=100,
                        unique=True,
                        verbose_name="slug",
                    ),
                ),
                (
                    "model_verbose_name",
                    models.CharField(
                        blank=True,
                        default=None,
                        editable=False,
                        max_length=255,
                        null=True,
                        verbose_name="Model verbose",
                    ),
                ),
                (
                    "model_name",
                    models.CharField(
                        blank=True,
                        default=None,
                        editable=False,
                        max_length=255,
                        null=True,
                        verbose_name="Model name",
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Comment",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        max_length=255,
                        null=True,
                        verbose_name="Field name",
                    ),
                ),
                (
                    "field_verbose_name",
                    models.CharField(
                        blank=True,
                        default=None,
                        editable=False,
                        max_length=255,
                        null=True,
                        verbose_name="Field verbose",
                    ),
                ),
                (
                    "meta",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        max_length=255,
                        null=True,
                        verbose_name="Field meta data",
                    ),
                ),
            ],
            options={
                "verbose_name": "User Tag",
                "verbose_name_plural": "User Tags",
                "ordering": ["model_verbose_name", "field_name", "tags"],
            },
        ),
        migrations.CreateModel(
            name="TaggedFieldModel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "model_name",
                    models.CharField(editable=False, max_length=128),
                ),
                (
                    "model_verbose_name",
                    models.CharField(editable=False, max_length=128),
                ),
                (
                    "field_name",
                    models.CharField(editable=False, max_length=128),
                ),
                (
                    "field_verbose_name",
                    models.CharField(editable=False, max_length=128),
                ),
                (
                    "tag_type",
                    models.CharField(
                        default="user",
                        editable=False,
                        help_text="This is the tag type, default is 'user'.",
                        max_length=20,
                        verbose_name="tags",
                    ),
                ),
                (
                    "default_tags",
                    models.CharField(
                        blank=True,
                        help_text="These are generated for each new user to get them started.",
                        max_length=255,
                        null=True,
                        verbose_name="Default tags for field.",
                    ),
                ),
                (
                    "content",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tagged Field Model",
                "verbose_name_plural": "Tagged Field Models",
            },
        ),
        migrations.AddConstraint(
            model_name="tagmesynchronise",
            constraint=models.UniqueConstraint(
                fields=("name",), name="unique_tag_synchronise_name"
            ),
        ),
        migrations.AddField(
            model_name="usertag",
            name="tagged_field",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="tagged_field",
                to="tag_me.taggedfieldmodel",
                verbose_name="Tagged Field",
            ),
        ),
        migrations.AddField(
            model_name="usertag",
            name="user",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="user_tags",
                to=settings.AUTH_USER_MODEL,
                verbose_name="User",
            ),
        ),
        migrations.AddConstraint(
            model_name="taggedfieldmodel",
            constraint=models.UniqueConstraint(
                fields=(
                    "content",
                    "model_name",
                    "model_verbose_name",
                    "field_name",
                    "field_verbose_name",
                ),
                name="unique_tagged_field_model",
            ),
        ),
        migrations.AddIndex(
            model_name="usertag",
            index=models.Index(
                fields=["user", "tagged_field", "tags"],
                name="tag_me_user_user_id_c18274_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="usertag",
            constraint=models.UniqueConstraint(
                fields=("user", "tagged_field"), name="unique_user_field_tag"
            ),
        ),
    ]



"""tag-me app custom form charfield."""

# from django.core import validators
from django.forms.fields import CharField
from django.utils.translation import gettext_lazy as _

from tag_me.utils.collections import FieldTagListFormatter


class TagMeCharField(CharField):
    """Custom tag-me Charfield Form

    Overrides to_python converting values into a
    :class: FieldTagListFormatter, returning the :method: toCSV() string
    in the format that :class: tag_me.utils.parser.parse_tags expects.
    """

    default_error_messages = {
        "invalid": _("Please enter a valid tag/tags"),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.formatter = FieldTagListFormatter()

    def to_python(self, value):
        """Return FieldTagListFormatter(value).toCSV() string."""

        self.formatter.clear()  # Ensure we start with an empty list
        self.formatter.add_tags(value)
        return self.formatter.toCSV(
            include_trailing_comma=True,  # Ensures correct tag string parsing
        )



"""tag-me model field tests"""

from typing import Dict

import pytest
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.test import SimpleTestCase
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra.django import TestCase

from tag_me.db.models.fields import TagMeCharField
from tag_me.models import UserTag
from tag_me.utils.collections import FieldTagListFormatter
from tests.models import TaggedFieldTestModel


class TestTagMeCharField(TestCase):
    def test_max_length_passed_to_formfield(self):
        """
        CharField passes its max_length attribute to form fields created using
        the formfield() method.

        Equivalent to Django test.
        """

        f1 = TagMeCharField()
        f1.model = UserTag()
        f2 = TagMeCharField(max_length=1234)
        f2.model = UserTag()

        assert 255 == f1.formfield().max_length
        assert 1234 == f2.formfield().max_length

    def test_emoji(self):
        """Equivalent to Django test."""
        tag = TaggedFieldTestModel.objects.create(
            tagged_field_1=["Smile 😀"]
            # slug="Whatever.",
        )
        tag.refresh_from_db()
        # tag-me tags are stored in the db with trailing comma to ensure they
        # are parsed correctly
        assert "Smile 😀," == tag.tagged_field_1

    # Removed issue-108. tag-me bypasses Django choices validation. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_assignment_from_choice_enum(self):
    #     """Equivalent to Django test."""
    #
    #     class Event(models.TextChoices):
    #         C = "Carnival!"
    #         F = "Festival!"
    #
    #     tag_1 = TaggedFieldTestModel.objects.create(
    #         tagged_field_1=Event.C,
    #         tagged_field_2=Event.F,
    #     )
    #     tag_1.refresh_from_db()
    #     assert "Carnival!" == tag_1.tagged_field_1
    #     assert "Festival!" == tag_1.tagged_field_2
    #     assert Event.C == tag_1.tagged_field_1
    #     assert Event.F == tag_1.tagged_field_2
    #     tag_2 = TaggedFieldTestModel.objects.get(
    #         tagged_field_1="Carnival!",
    #     )
    #     assert tag_2 == tag_1
    #     assert Event.C == tag_2.tagged_field_1


class TestMethods(SimpleTestCase):
    """Equivalent to Django test."""

    def test_deconstruct(self):
        f = TagMeCharField()
        *_, kwargs = f.deconstruct()
        assert {"max_length": 255} == kwargs
        f = TagMeCharField(db_collation="utf8_esperanto_ci")
        *_, kwargs = f.deconstruct()
        assert {
            "db_collation": "utf8_esperanto_ci",
            "max_length": 255,
        } == kwargs


class TestValidation(SimpleTestCase):
    class Choices(models.TextChoices):
        A = "a", "A"
        B = "b", "B"
        C = "c", "C"

    def test_validators(self):
        f = TagMeCharField(
            max_length=1234,
        )

        assert any(
            x
            for x in f.validators
            if isinstance(x, validators.MaxLengthValidator)
        )

    def test_charfield_raises_error_on_empty_string(self):
        """Equivelant to Django test."""
        f = TagMeCharField()
        msg = "This field cannot be blank."
        with self.assertRaisesMessage(ValidationError, msg):
            f.clean("", None)

    def test_charfield_cleans_empty_string_when_blank_true(self):
        f = TagMeCharField(blank=True)
        assert f.clean("", None) == ""

    # Removed issue-108. tag-me bypasses Django choices validation. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_charfield_with_choices_cleans_valid_choice(self):
    #     f = TagMeCharField(max_length=1, choices=[("a", "A"), ("b", "B")])
    #     assert f.clean("a", None) == "a"

    # Removed issue-108. tag-me does not use enum in choices. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_charfield_with_choices_raises_error_on_invalid_choice(self):
    #     f = TagMeCharField(choices=[("a", "A"), ("b", "B")])
    #     msg = "Value 'not a' is not a valid choice."
    #     with self.assertRaisesMessage(ValidationError, msg):
    #         f.clean("not a", None)

    # Removed issue-108. tag-me does not use enum in choices. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_enum_choices_cleans_valid_string(self):
    #     f = TagMeCharField(choices=self.Choices.choices, max_length=1)
    #     assert f.clean("c", None) == "c"

    # Removed issue-108. tag-me does not use enum in choices. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_enum_choices_invalid_input(self):
    #     f = TagMeCharField(choices=self.Choices.choices, max_length=1)
    #     msg = "Value 'd' is not a valid choice."
    #     with self.assertRaisesMessage(ValidationError, msg):
    #         f.clean("d", None)
    #

    def test_charfield_raises_error_on_empty_input(self):
        f = TagMeCharField(null=False)
        # msg = "This field cannot be null."  This message fails.
        msg = "This field cannot be blank."
        with self.assertRaisesMessage(ValidationError, msg):
            f.clean(None, None)

    def test_charfield_not_editable(self):
        f = TagMeCharField(editable=False)
        # All tag-me tags are stored with a trailing comma to ensure correct
        # parsing to and from db
        value = "testing,"

        assert f.clean(value, None) == value

    # Removed issue-108. tag-me bypasses Django choices validation. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_charfield_choice_not_valid_and_not_in_empty(self):
    #     f = TagMeCharField(choices=self.Choices.choices)
    #     value = "invalid"
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean(value, None)
    #     assert "Value 'invalid' is not a valid choice." in str(exc.value)
    #     assert exc.type == ValidationError

    # Removed issue-108. tag-me bypasses Django choices validation. The choices
    # are added to a FieldTagListFormatter instance.
    # def test_charfield_choice_in_empty_values(self):
    #     f = TagMeCharField(choices=self.Choices.choices)
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean(None, None)
    #     assert "This field cannot be null." in str(exc.value)
    #     assert exc.type == ValidationError
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean("", None)
    #     assert "This field cannot be blank." in str(exc.value)
    #     assert exc.type == ValidationError
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean([], None)
    #     assert "This field cannot be blank." in str(exc.value)
    #     assert exc.type == ValidationError
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean((), None)
    #     assert "This field cannot be blank." in str(exc.value)
    #     assert exc.type == ValidationError
    #     with pytest.raises(ValidationError) as exc:
    #         f.clean({}, None)
    #     assert "This field cannot be blank." in str(exc.value)
    #     assert exc.type == ValidationError


class TestTagMeCharfieldtoPython(SimpleTestCase):
    """Test passing various in formats and correct conversion is returned.

    We don't use hypothesis here because the inputs and outputs must follow
    a specific format.
    """

    def test_tags_as_str_input(self):
        test_str1 = "apple ball cat"
        test_str2 = "apple,ball cat"
        test_str3 = '"apple, ball" cat dog'
        test_str4 = '"apple, ball", cat dog'
        test_str5 = 'apple "ball cat" dog'
        test_str6 = '"apple""ball dog'
        f = TagMeCharField()

        assert f.to_python(test_str1) == FieldTagListFormatter(
            test_str1
        ).toCSV(
            include_trailing_comma=True,
        )

        assert f.to_python(test_str2) == FieldTagListFormatter(
            test_str2
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_str3) == FieldTagListFormatter(
            test_str3
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_str4) == FieldTagListFormatter(
            test_str4
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_str5) == FieldTagListFormatter(
            test_str5
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_str6) == FieldTagListFormatter(
            test_str6
        ).toCSV(
            include_trailing_comma=True,
        )

    def test_tags_as_list_input(self):
        test_lst1 = ["apple ball cat"]
        test_lst2 = ["apple,ball cat"]
        test_lst3 = ['"apple, ball" cat dog']
        test_lst4 = ['"apple, ball", cat dog']
        test_lst5 = ['apple "ball cat" dog']
        test_lst6 = ['"apple""ball dog']
        f = TagMeCharField()

        assert f.to_python(test_lst1) == FieldTagListFormatter(
            test_lst1
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_lst2) == FieldTagListFormatter(
            test_lst2
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_lst3) == FieldTagListFormatter(
            test_lst3
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_lst4) == FieldTagListFormatter(
            test_lst4
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_lst5) == FieldTagListFormatter(
            test_lst5
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_lst6) == FieldTagListFormatter(
            test_lst6
        ).toCSV(
            include_trailing_comma=True,
        )

    def test_tags_as_dict_input(self):
        test_dict1 = {"tags": ["apple ball cat"]}
        test_dict2 = {"tags": ["apple,ball cat"]}
        test_dict3 = {"tags": ['"apple, ball" cat dog']}
        test_dict4 = {"tags": ['"apple, ball", cat dog']}
        test_dict5 = {"tags": ['apple "ball cat" dog']}
        test_dict6 = {"tags": ['"apple""ball dog']}
        f = TagMeCharField()

        assert f.to_python(test_dict1) == FieldTagListFormatter(
            test_dict1
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_dict2) == FieldTagListFormatter(
            test_dict2
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_dict3) == FieldTagListFormatter(
            test_dict3
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_dict4) == FieldTagListFormatter(
            test_dict4
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_dict5) == FieldTagListFormatter(
            test_dict5
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_dict6) == FieldTagListFormatter(
            test_dict6
        ).toCSV(
            include_trailing_comma=True,
        )

    def test_tags_as_set_input(self):
        test_set1 = {"apple ball cat"}
        test_set2 = {"apple,ball cat"}
        test_set3 = {'"apple, ball" cat dog'}
        test_set4 = {'"apple, ball", cat dog'}
        test_set5 = {'apple "ball cat" dog'}
        test_set6 = {'"apple""ball dog'}
        f = TagMeCharField()

        assert f.to_python(test_set1) == FieldTagListFormatter(
            test_set1
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_set2) == FieldTagListFormatter(
            test_set2
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_set3) == FieldTagListFormatter(
            test_set3
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_set4) == FieldTagListFormatter(
            test_set4
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_set5) == FieldTagListFormatter(
            test_set5
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_set6) == FieldTagListFormatter(
            test_set6
        ).toCSV(
            include_trailing_comma=True,
        )

    # ..todo:: probably a duplicate of test in test_collections. Review in refactor
    # def test_tags_as_unsupported_input(self):
    #     test_uns1 = 5
    #     test_uns2 = 1.2
    #     test_uns3 = True
    #     f = TagMeCharField()
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.to_python(test_uns1)
    #
    #         assert "['5 is not type str, set, list or dict']" in str(
    #             exc.value
    #         ).replace("\\", "")
    #         assert exc.type == ValidationError
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.to_python(test_uns2)
    #
    #         assert "['1.2 is not type str, set, list or dict']" in str(
    #             exc.value
    #         ).replace("\\", "")
    #         assert exc.type == ValidationError
    #
    #     with pytest.raises(ValidationError) as exc:
    #         f.to_python(test_uns3)
    #
    #         assert "['True is not type str, set, list or dict']" in str(
    #             exc.value
    #         ).replace("\\", "")
    #         assert exc.type == ValidationError

    def test_tags_input_is_none(self):
        test_none = None
        f = TagMeCharField()
        assert f.to_python(test_none) == ""

    def test_tags_input_is_FieldTagListFormatter(self):
        test_ftf1 = FieldTagListFormatter({"apple ball cat"})
        test_ftf2 = FieldTagListFormatter({"apple,ball cat"})
        test_ftf3 = FieldTagListFormatter({'"apple, ball" cat dog'})
        test_ftf4 = FieldTagListFormatter({'"apple, ball", cat dog'})
        test_ftf5 = FieldTagListFormatter({'apple "ball cat" dog'})
        test_ftf6 = FieldTagListFormatter({'"apple""ball dog'})
        f = TagMeCharField()

        assert f.to_python(test_ftf1) == FieldTagListFormatter(
            test_ftf1
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_ftf2) == FieldTagListFormatter(
            test_ftf2
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_ftf3) == FieldTagListFormatter(
            test_ftf3
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_ftf4) == FieldTagListFormatter(
            test_ftf4
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_ftf5) == FieldTagListFormatter(
            test_ftf5
        ).toCSV(
            include_trailing_comma=True,
        )
        assert f.to_python(test_ftf6) == FieldTagListFormatter(
            test_ftf6
        ).toCSV(
            include_trailing_comma=True,
        )

    def test_tags_input_is_choices(self):
        class Event(models.TextChoices):
            C = "Carnival!"
            F = "Festival!"

        f = TagMeCharField(choices=Event.choices)

        # Check internal representation of choices
        assert f._tag_choices == ["Carnival!", "Festival!"]

        # Check the choices are formatted and saved to the db correctly
        assert f.to_python(f._tag_choices) == "Carnival!, Festival!,"

        # Check Django choices machinery is disabled.
        assert f.choices is None

    def test_tags_input_is_null_only_tags(self):
        null_tags = [
            "null",
            "Null",
            "NULL",
        ]

        f = TagMeCharField()

        assert f.to_python(null_tags) == ""

    # ..todo:: probably a duplicate of test in test_collections. Review in refactor
    # def test_tags_input_includes_null_tags(self):
    #     null_tags = [
    #         "null",
    #         "Null",
    #         "NULL",
    #         "Not Null",
    #         "Still Not Null",
    #     ]
    #
    #     f = TagMeCharField()
    #
    #     assert f.to_python(null_tags) == "Not Null, Still Not Null"


class TestTagMeCharFieldOtherMethods(TestCase):
    """Test other overridden Charfield Methods"""

    def test_default_form_field_form_class_widget(self):
        f = TagMeCharField()
        f.model = UserTag()
        assert (
            str(type(f.formfield()))
            == "<class 'tag_me.db.forms.fields.TagMeCharField'>"
        )

    def test_admin_form_field_form_class_widget(self):
        kwargs: dict[str, str] = {"widget": "django.contrib.admin.widgets"}

        f = TagMeCharField()
        f.model = UserTag()

        assert (
            str(type(f.formfield(**kwargs)))
            == "<class 'django.forms.fields.CharField'>"
        )



"""tests Models file."""

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import pgettext_lazy as _

from tag_me.db.models.fields import TagMeCharField
from tag_me.models import TagBase

User = get_user_model()


class TagTestBase(TagBase):
    """A minimal tag to test `TagBase ABC`."""


class TaggedFieldTestModel(TagTestBase):
    """A model for testing the Tag base class"""

    class Meta:
        verbose_name = _(
            "A verbose name",
            "Tagged Field Test Model",
        )
        verbose_name_plural = _(
            "A verbose name",
            "Tagged Field Test Models",
        )

    tagged_field_1 = TagMeCharField(
        max_length=255,
        blank=True,
        null=True,
    )
    tagged_field_2 = TagMeCharField(
        max_length=255,
        blank=True,
        null=True,
    )


class Post(models.Model):
    title = TagMeCharField(
        max_length=255,
        blank=True,
        null=True,
    )
    body = models.TextField()
